/**
 * Copyright (c) 2011 panStamp <contact@panstamp.com>
 * 
 * This file is part of the panStamp project.
 * 
 * panStamp  is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 * 
 * panStamp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with panStamp; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 
 * USA
 * 
 * Author: Daniel Berenguer
 * Creation date: 03/03/2011
 */

#include "panstamp.h"
#include "nvolat.h"
#include <util/delay.h>
 
/**
 * Macros
 */
// Select (SPI) CC1101
#define cc1101_Select()  bitClear(PORT_SPI_SS, BIT_SPI_SS)
// Deselect (SPI) CC1101
#define cc1101_Deselect()  bitSet(PORT_SPI_SS, BIT_SPI_SS)
// Wait until SPI MISO line goes low
#define wait_Miso()  while(bitRead(PORT_SPI_MISO, BIT_SPI_MISO))
// Get GDO0 pin state
#define getGDO0state()  bitRead(PORT_GDO0, BIT_GDO0)
// Wait until GDO0 line goes high
#define wait_GDO0_high()  while(!getGDO0state())
// Wait until GDO0 line goes low
#define wait_GDO0_low()  while(getGDO0state())

 /**
  * PATABLE
  */
//const byte paTable[8] = {0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60};

/**
 * CC1101
 * 
 * Class constructor
 */
/*CC1101::CC1101(void)
{
  paTableByte = PA_LowPower;            // Priority = Low power
}*/


/**
 * wakeUp
 * 
 * Wake up CC1101 from Power Down state
 */
void wakeUpCC1101(void)
{
  cc1101_Select();                      // Select CC1101
  wait_Miso();                          // Wait until MISO goes low
  cc1101_Deselect();                    // Deselect CC1101
}

/**
 * writeReg
 * 
 * Write single register into the CC1101 IC via SPI
 * 
 * 'regAddr'	Register address
 * 'value'	Value to be writen
 */
void writeRegCC1101(byte regAddr, byte value) 
{ 

  cc1101_Select();                      // Select CC1101
  wait_Miso();                          // Wait until MISO goes low
  sendspi(regAddr);                     //spi.send(regAddr); // Send register address
  //testdebug     ---------------------------------------------------------------------------------------------------------
   

  sendspi(value);                       // Send value
  cc1101_Deselect();                    // Deselect CC1101
}

/**
 * writeBurstReg
 * 
 * Write multiple registers into the CC1101 IC via SPI
 * 
 * 'regAddr'	Register address
 * 'buffer'	Data to be writen
 * 'len'	Data length
 */
void writeBurstRegCC1101(byte regAddr, byte* buffer, byte len)
{
  byte addr, i;
  
  addr = regAddr | WRITE_BURST;         // Enable burst transfer
  cc1101_Select();                      // Select CC1101
  wait_Miso();                          // Wait until MISO goes low
  sendspi(addr);                        // Send register address
  
  for(i=0 ; i<len ; i++)
    sendspi(buffer[i]);                 // Send value

  cc1101_Deselect();                    // Deselect CC1101  
}

/**
 * cmdStrobeCC1101cmdStrobe
 * 
 * Send command strobe to the CC1101 IC via SPI
 * 
 * 'cmd'	Command strobe
 */     
void cmdStrobeCC1101(byte cmd) 
{
  cc1101_Select();                      // Select CC1101
  wait_Miso();                          // Wait until MISO goes low
  sendspi(cmd);                         // Send strobe command
  cc1101_Deselect();                    // Deselect CC1101
}

/**
 * readReg
 * 
 * Read CC1101 register via SPI
 * 
 * 'regAddr'	Register address
 * 'regType'	Type of register: CC1101_CONFIG_REGISTER or CC1101_STATUS_REGISTER
 * 
 * Return:
 * 	Data byte returned by the CC1101 IC
 */
byte readRegCC1101(byte regAddr, byte regType) 
{
  byte addr, val;

  addr = regAddr | regType;
  cc1101_Select();                      // Select CC1101
  wait_Miso();                          // Wait until MISO goes low
  sendspi(addr);                        // Send register address
  val = sendspi(0x00);                  // Read result
  cc1101_Deselect();                    // Deselect CC1101

  return val;
}

/**
 * readBurstReg
 * 
 * Read burst data from CC1101 via SPI
 * 
 * 'buffer'	Buffer where to copy the result to
 * 'regAddr'	Register address
 * 'len'	Data length
 */
void readBurstRegCC1101(byte * buffer, byte regAddr, byte len) 
{
  byte addr, i;
  
  addr = regAddr | READ_BURST;
  cc1101_Select();                      // Select CC1101
  wait_Miso();                          // Wait until MISO goes low
  sendspi(addr);                        // Send register address
  for(i=0 ; i<len ; i++)
    buffer[i] = sendspi(0x00);         // Read result byte by byte
  cc1101_Deselect();                    // Deselect CC1101
}

/**
 * reset
 * 
 * Reset CC1101
 */
void resetCC1101(void) 
{
  cc1101_Deselect();                    // Deselect CC1101
  _delay_us(5);                  //delayMicroseconds(5);
  cc1101_Select();                      // Select CC1101
  _delay_us(10);                 //delayMicroseconds(10);
  cc1101_Deselect();                    // Deselect CC1101
  _delay_us(41);                 //delayMicroseconds(41);
  cc1101_Select();                      // Select CC1101

  wait_Miso();                          // Wait until MISO goes low
  sendspi(CC1101_SRES);                 // Send reset command strobe
  wait_Miso();                          // Wait until MISO goes low

  cc1101_Deselect();                    // Deselect CC1101

  setDefaultRegsCC1101();               // Reconfigure CC1101
//  setRegsFromEeprom();                // Take user settings from EEPROM
}

/**
 * setDefaultRegs
 * 
 * Configure CC1101 registers
 */
void setDefaultRegsCC1101(void) 
{
  boolean save=false;

  writeRegCC1101(CC1101_IOCFG2,  CC1101_DEFVAL_IOCFG2);
  writeRegCC1101(CC1101_IOCFG1,  CC1101_DEFVAL_IOCFG1);
  writeRegCC1101(CC1101_IOCFG0,  CC1101_DEFVAL_IOCFG0);
  writeRegCC1101(CC1101_FIFOTHR,  CC1101_DEFVAL_FIFOTHR);
  writeRegCC1101(CC1101_PKTLEN,  CC1101_DEFVAL_PKTLEN);
  writeRegCC1101(CC1101_PKTCTRL1,  CC1101_DEFVAL_PKTCTRL1);
  writeRegCC1101(CC1101_PKTCTRL0,  CC1101_DEFVAL_PKTCTRL0);
  
  // Sync word
  if ( ((panstamp.radio.syncWord[0] == 0xFF) && (panstamp.radio.syncWord[1] == 0xFF)) || (panstamp.radio.syncWord[0] == 0) || (panstamp.radio.syncWord[1] == 0))  //modif
    setSyncWordCC1101_n3(CC1101_DEFVAL_SYNC1, CC1101_DEFVAL_SYNC0, save); // Set default synchronization word
  else
    setSyncWordCC1101_n2(panstamp.radio.syncWord, save);   // a voir

  // Device addres
  if (panstamp.radio.devAddress == 0 || panstamp.radio.devAddress == 0xFF)
    setDevAddressCC1101(CC1101_DEFVAL_ADDR,save); // Set default device address
  else
    writeRegCC1101(CC1101_ADDR, panstamp.radio.devAddress);

  // Set default frequency channel
  if (panstamp.radio.channel == 0xFF)
    setChannelCC1101(CC1101_DEFVAL_CHANNR,save);    // Default value
  else
    setChannelCC1101(panstamp.radio.channel,save);
  
  writeRegCC1101(CC1101_FSCTRL1,  CC1101_DEFVAL_FSCTRL1);
  writeRegCC1101(CC1101_FSCTRL0,  CC1101_DEFVAL_FSCTRL0);

  // Carrier frequency
  setCarrierFreqCC1101(panstamp.radio.carrierFreq);

  writeRegCC1101(CC1101_MDMCFG4,  CC1101_DEFVAL_MDMCFG4);
  writeRegCC1101(CC1101_MDMCFG3,  CC1101_DEFVAL_MDMCFG3);
  writeRegCC1101(CC1101_MDMCFG2,  CC1101_DEFVAL_MDMCFG2);
  writeRegCC1101(CC1101_MDMCFG1,  CC1101_DEFVAL_MDMCFG1);
  writeRegCC1101(CC1101_MDMCFG0,  CC1101_DEFVAL_MDMCFG0);
  writeRegCC1101(CC1101_DEVIATN,  CC1101_DEFVAL_DEVIATN);
  writeRegCC1101(CC1101_MCSM2,  CC1101_DEFVAL_MCSM2);
  writeRegCC1101(CC1101_MCSM1,  CC1101_DEFVAL_MCSM1);
  writeRegCC1101(CC1101_MCSM0,  CC1101_DEFVAL_MCSM0);
  writeRegCC1101(CC1101_FOCCFG,  CC1101_DEFVAL_FOCCFG);
  writeRegCC1101(CC1101_BSCFG,  CC1101_DEFVAL_BSCFG);
  writeRegCC1101(CC1101_AGCCTRL2,  CC1101_DEFVAL_AGCCTRL2);
  writeRegCC1101(CC1101_AGCCTRL1,  CC1101_DEFVAL_AGCCTRL1);
  writeRegCC1101(CC1101_AGCCTRL0,  CC1101_DEFVAL_AGCCTRL0);
  writeRegCC1101(CC1101_WOREVT1,  CC1101_DEFVAL_WOREVT1);
  writeRegCC1101(CC1101_WOREVT0,  CC1101_DEFVAL_WOREVT0);
  writeRegCC1101(CC1101_WORCTRL,  CC1101_DEFVAL_WORCTRL);
  writeRegCC1101(CC1101_FREND1,  CC1101_DEFVAL_FREND1);
  writeRegCC1101(CC1101_FREND0,  CC1101_DEFVAL_FREND0);
  writeRegCC1101(CC1101_FSCAL3,  CC1101_DEFVAL_FSCAL3);
  writeRegCC1101(CC1101_FSCAL2,  CC1101_DEFVAL_FSCAL2);
  writeRegCC1101(CC1101_FSCAL1,  CC1101_DEFVAL_FSCAL1);
  writeRegCC1101(CC1101_FSCAL0,  CC1101_DEFVAL_FSCAL0);
  writeRegCC1101(CC1101_RCCTRL1,  CC1101_DEFVAL_RCCTRL1);
  writeRegCC1101(CC1101_RCCTRL0,  CC1101_DEFVAL_RCCTRL0);
  writeRegCC1101(CC1101_FSTEST,  CC1101_DEFVAL_FSTEST);
  writeRegCC1101(CC1101_PTEST,  CC1101_DEFVAL_PTEST);
  writeRegCC1101(CC1101_AGCTEST,  CC1101_DEFVAL_AGCTEST);
  writeRegCC1101(CC1101_TEST2,  CC1101_DEFVAL_TEST2);
  writeRegCC1101(CC1101_TEST1,  CC1101_DEFVAL_TEST1);
  writeRegCC1101(CC1101_TEST0,  CC1101_DEFVAL_TEST0);
}

/**
 * init
 * 
 * Initialize CC1101
 * 
 * 'freq'	New carrier frequency
 */
void initCC1101(byte freq) 
{
  freq=CFREQ_868;
  panstamp.radio.carrierFreq = freq;
  
  initspi();                           // Initialize SPI interface
  DDRD &=~(_BV(PD2));//cbi(DDRD,PD2);                       // PD2 en entree
  resetCC1101();                       // Reset CC1101

  // Configure PATABLE
  //writeBurstReg(CC1101_PATABLE, (byte*)paTable, 8);
  writeRegCC1101(CC1101_PATABLE, panstamp.radio.paTableByte);
}

/**
 * setSyncWord
 * 
 * Set synchronization word
 * 
 * 'syncH'	Synchronization word - High byte
 * 'syncL'	Synchronization word - Low byte
 * 'save' If TRUE, save parameter in EEPROM
 */
void setSyncWordCC1101_n3(uint8_t syncH, uint8_t syncL, bool save) 
{
  save=false;

 /* if ((panstamp.radio.syncWord[0] != syncH) || (panstamp.radio.syncWord[1] != syncL))
  {
    writeRegCC1101(CC1101_SYNC1, syncH);
    writeRegCC1101(CC1101_SYNC0, syncL);
    panstamp.radio.syncWord[0] = syncH;
    panstamp.radio.syncWord[1] = syncL;
    // Save in EEPROM
    if (save)
    {
      writeEEPROM(EEPROM_SYNC_WORD, syncH);
      writeEEPROM(EEPROM_SYNC_WORD + 1, syncL);
    }
  }*/

   writeRegCC1101(CC1101_SYNC1, syncH);
   writeRegCC1101(CC1101_SYNC0, syncL); 
    //test debug;
    //sbi(DDRD,PD3);
    //sbi(PORTD,PD3);
   //_delay_ms(1000);
   panstamp.radio.syncWord[0] = syncH;
   panstamp.radio.syncWord[1] = syncL;
  
   // cbi(PORTD,PD3);
   // _delay_ms(1000);
   // sbi(PORTD,PD3);
   //_delay_ms(1000);
}

/**
 * setSyncWord (overriding method)
 * 
 * Set synchronization word
 * 
 * 'syncH'	Synchronization word - pointer to 2-byte array
 * 'save' If TRUE, save parameter in EEPROM
 */
void setSyncWordCC1101_n2(byte *sync, bool save) 
{
  save=false;
  setSyncWordCC1101_n3(sync[0], sync[1], save);
}

/**
 * setDevAddress
 * 
 * Set device address
 * 
 * 'addr'	Device address
 * 'save' If TRUE, save parameter in EEPROM
 */
void setDevAddressCC1101(byte addr, bool save) 
{
  save=false;

  if ((panstamp.radio.devAddress )!= addr)
  {
   writeRegCC1101(CC1101_ADDR, addr);
    panstamp.radio.devAddress = addr;
    // Save in EEPROM
    if (save)
      writeEEPROM(EEPROM_DEVICE_ADDR+1, addr);  
  }
}

/**
 * setChannel
 * 
 * Set frequency channel
 * 
 * 'chnl'	Frequency channel
 * 'save' If TRUE, save parameter in EEPROM
 */
void setChannelCC1101(byte chnl, bool save) 
{
  save=false;
  if ( (panstamp.radio.channel)!= chnl)
  {
    writeRegCC1101(CC1101_CHANNR,  chnl);
    panstamp.radio.channel = chnl;
    // Save in EEPROM
    if (save)
      writeEEPROM(EEPROM_FREQ_CHANNEL, chnl);
  }
}

/**
 * setCarrierFreq
 * 
 * Set carrier frequency
 * 
 * 'freq'	New carrier frequency
 */
void setCarrierFreqCC1101(byte freq)
{
  switch(freq)
  {
    case CFREQ_915:
      writeRegCC1101(CC1101_FREQ2,  CC1101_DEFVAL_FREQ2_915);
      writeRegCC1101(CC1101_FREQ1,  CC1101_DEFVAL_FREQ1_915);
      writeRegCC1101(CC1101_FREQ0,  CC1101_DEFVAL_FREQ0_915);
      break;
    case CFREQ_433:
      writeRegCC1101(CC1101_FREQ2,  CC1101_DEFVAL_FREQ2_433);
      writeRegCC1101(CC1101_FREQ1,  CC1101_DEFVAL_FREQ1_433);
      writeRegCC1101(CC1101_FREQ0,  CC1101_DEFVAL_FREQ0_433);
      break;
    case CFREQ_918:
      writeRegCC1101(CC1101_FREQ2,  CC1101_DEFVAL_FREQ2_918);
      writeRegCC1101(CC1101_FREQ1,  CC1101_DEFVAL_FREQ1_918);
      writeRegCC1101(CC1101_FREQ0,  CC1101_DEFVAL_FREQ0_918);
      break;
    default:
      writeRegCC1101(CC1101_FREQ2,  CC1101_DEFVAL_FREQ2_868);
      writeRegCC1101(CC1101_FREQ1,  CC1101_DEFVAL_FREQ1_868);
      writeRegCC1101(CC1101_FREQ0,  CC1101_DEFVAL_FREQ0_868);
      break;
  }
   
  panstamp.radio.carrierFreq = freq;  
}

/**
 * setRegsFromEeprom
 * 
 * Set registers from EEPROM
 */
/*
void CC1101::setRegsFromEeprom(void)
{
  byte bVal;
  byte arrV[2];

  // Read RF channel from EEPROM
  bVal = EEPROM.read(EEPROM_FREQ_CHANNEL);
  // Set RF channel
  if (bVal < NUMBER_OF_FCHANNELS )
    setChannel(bVal, false);
  // Read Sync word from EEPROM
  arrV[0] = EEPROM.read(EEPROM_SYNC_WORD);
  arrV[1] = EEPROM.read(EEPROM_SYNC_WORD + 1);
  // Set Sync word. 0x00 and 0xFF values are not allowed
  if (((arrV[0] != 0x00) && (arrV[0] != 0xFF)) || ((arrV[1] != 0x00) && (arrV[1] != 0xFF)))
    setSyncWord(arrV[0], arrV[1], false);
  // Read device address from EEPROM
  bVal = EEPROM.read(EEPROM_DEVICE_ADDR+1);
  // Set device address
  if (bVal > 0)
    setDevAddress(bVal, false);
}
*/
/**
 * setPowerDownState
 * 
 * Put CC1101 into power-down state
 */
void setPowerDownStateCC1101() 
{
  // Comming from RX state, we need to enter the IDLE state first
  cmdStrobeCC1101(CC1101_SIDLE);
  // Enter Power-down state
  cmdStrobeCC1101(CC1101_SPWD);

}

/**
 * sendData
 * 
 * Send data packet via RF
 * 
 * 'packet'	Packet to be transmitted. First byte is the destination address
 *
 *  Return:
 *    True if the transmission succeeds
 *    False otherwise
 */
boolean sendDataCC1101(CCPACKET packet)
{
  byte marcState;
  bool res = false;
 
  // Declare to be in Tx state. This will avoid receiving packets whilst
  // transmitting
  panstamp.radio.rfState = RFSTATE_TX;

  // Enter RX state
  setRxState();

  // Check that the RX state has been entered
  while (((marcState = readStatusReg(CC1101_MARCSTATE)) & 0x1F) != 0x0D)
  {
    if (marcState == 0x11)        // RX_OVERFLOW
      flushRxFifo();              // flush receive queue
  }

    _delay_us(490);               //delayMicroseconds(500);

  // Set data length at the first position of the TX FIFO
  writeRegCC1101(CC1101_TXFIFO,  packet.length);
  // Write data into the TX FIFO
  writeBurstRegCC1101(CC1101_TXFIFO, packet.data, packet.length);

  // CCA enabled: will enter TX state only if the channel is clear
  setTxState();

  // Check that TX state is being entered (state = RXTX_SETTLING)
  marcState = readStatusReg(CC1101_MARCSTATE) & 0x1F;
  if((marcState != 0x13) && (marcState != 0x14) && (marcState != 0x15))
  {
    setIdleState();       // Enter IDLE state
    flushTxFifo();        // Flush Tx FIFO
    setRxState();         // Back to RX state

    // Declare to be in Rx state
    panstamp.radio.rfState = RFSTATE_RX;
    return false;
  }

  // Wait for the sync word to be transmitted
  wait_GDO0_high();

  // Wait until the end of the packet transmission
  wait_GDO0_low();

  // Check that the TX FIFO is empty
  if((readStatusReg(CC1101_TXBYTES) & 0x7F) == 0)
    res = true;

  setIdleState();       // Enter IDLE state
  flushTxFifo();        // Flush Tx FIFO

  // Enter back into RX state
  setRxState();

  // Declare to be in Rx state
  panstamp.radio.rfState = RFSTATE_RX;

  return res;
}

/**
 * receiveData
 * 
 * Read data packet from RX FIFO
 *
 * 'packet'	Container for the packet received
 * 
 * Return:
 * 	Amount of bytes received
 */
byte receiveDataCC1101(CCPACKET * packet)
{
  byte val;
  byte rxBytes = readStatusReg(CC1101_RXBYTES);

  // Any byte waiting to be read and no overflow?
  if (rxBytes & 0x7F && !(rxBytes & 0x80))
  {
    // Read data length
    packet->length = readConfigReg(CC1101_RXFIFO);
    // If packet is too long
    if (packet->length > CC1101_DATA_LEN)
      packet->length = 0;   // Discard packet
    else
    {
      // Read data packet
      readBurstRegCC1101(packet->data, CC1101_RXFIFO, packet->length);
      // Read RSSI
      packet->rssi = readConfigReg(CC1101_RXFIFO);
      // Read LQI and CRC_OK
      val = readConfigReg(CC1101_RXFIFO);
      packet->lqi = val & 0x7F;
      packet->crc_ok = bitRead(val, 7);
    }
  }
  else
    packet->length = 0;

  setIdleState();       // Enter IDLE state
  flushRxFifo();        // Flush Rx FIFO
  //cmdStrobe(CC1101_SCAL);

  // Back to RX state
  setRxState();

  return packet->length;
}
